/* GPLv2 applies
 * SVN revision: $Revision: 589 $
 * (C) 2006-2013 by folkert@vanheusden.com
 */
#define _GNU_SOURCE
#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <libgen.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <ncursesw/ncurses.h>
#include <ncursesw/panel.h>

#include "gen.h"
#include "error.h"
#include "term.h"
#include "buffer.h"
#include "channels.h"
#include "servers.h"
#include "utils.h"
#include "irc.h"
#include "loop.h"
#include "grep_filter.h"
#include "main.h"
#include "wordcloud.h"
#include "dcc.h"
#include "theme.h"
#include "config.h"
#include "colors.h"

BOOL topic_scroll = FALSE;
BOOL partial_highlight_match = FALSE;
const char *part_message = NULL, *server_exit_message = NULL;
int check_for_mail = 0;
int delay_before_reconnect = 3;
int max_channel_record_lines;
BOOL auto_private_channel;
BOOL highlight = FALSE, fuzzy_highlight = FALSE;
BOOL notice_in_server_channel = FALSE;
BOOL colors_meta = TRUE, colors_all = TRUE;
BOOL allow_invite = TRUE;
BOOL store_config_on_exit = FALSE;
BOOL show_parts = TRUE;
BOOL show_mode_changes = TRUE;
BOOL show_nick_change = TRUE;
BOOL show_joins = TRUE;
BOOL auto_rejoin = FALSE;
BOOL nick_color = FALSE;
BOOL use_nonbasic_colors = FALSE;
char *conf_file = NULL;
const char *finger_str = NULL;
BOOL inverse_window_heading = TRUE;
char *ignore_file = NULL;
ignore *ignores = NULL;
int n_ignores = 0;
BOOL remember_channels = TRUE;
BOOL mark_personal_messages = FALSE;
BOOL update_clock_at_data = FALSE;
BOOL irc_keepalive = TRUE;
BOOL auto_reconnect = TRUE;
BOOL space_after_start_marker = FALSE;
int nick_sleep = 4;

favorite *favorite_channels = NULL;
int n_favorite_channels = 0, favorite_channels_index = 0;

cnf_entry cnf_pars[] = {
	{ "enable colors", CNF_BOOL, &colors_all },
	{ "use non basic colors", CNF_BOOL, &use_nonbasic_colors },
	{ "nick color", CNF_BOOL, &nick_color },
	{ "meta-texts colors", CNF_BOOL, &colors_meta },
	{ "default color pair", CNF_COLOR, &default_colorpair },
	{ "highlight color pair", CNF_COLOR, &highlight_colorpair },
	{ "highlight", CNF_BOOL, &highlight },
	{ "fuzzy highlight match", CNF_BOOL, &fuzzy_highlight },
	{ "meta color pair", CNF_COLOR, &meta_colorpair },
	{ "error color pair", CNF_COLOR, &error_colorpair },
	{ "temp color pair", CNF_COLOR, &temp_colorpair },
	{ "inverse window heading", CNF_BOOL, &inverse_window_heading },
	{ "partial highlight match", CNF_BOOL, &partial_highlight_match },
	{ "automatically create private channel", CNF_BOOL, &auto_private_channel },
	{ "DCC path", CNF_STRING, &dcc_path },
	{ "part message", CNF_STRING, &part_message },
	{ "leave a server message", CNF_STRING, &server_exit_message },
	{ "ignores file", CNF_STRING, &ignore_file },
	{ "channel buffering", CNF_VALUE, &max_channel_record_lines },
	{ "delay before connect", CNF_VALUE, &delay_before_reconnect },
	{ "scroll topic", CNF_BOOL, &topic_scroll },
	{ "notices in server window", CNF_BOOL, &notice_in_server_channel },
	{ "allow invites", CNF_BOOL, &allow_invite },
	{ "store config. on exit", CNF_BOOL, &store_config_on_exit },
	{ "show PARTs", CNF_BOOL, &show_parts },
	{ "show mode changes", CNF_BOOL, &show_mode_changes },
	{ "show nick changes", CNF_BOOL, &show_nick_change },
	{ "show JOINs/QUITs", CNF_BOOL, &show_joins },
	{ "automatically re-join after kick", CNF_BOOL, &auto_rejoin },
	{ "mark personal messages-channels", CNF_BOOL, &mark_personal_messages },
	{ "only update clock when data comes in", CNF_BOOL, &update_clock_at_data },
	{ "IRC ping (detect off-line sooner)", CNF_BOOL, &irc_keepalive },
	{ "auto reconnect when connection lost", CNF_BOOL, &auto_reconnect },
	{ "space between \">\" and input line", CNF_BOOL, &space_after_start_marker },
	{ NULL, 0, NULL }
	};

BOOL load_ignore_list(const char *file)
{
	int fd = open(file, O_RDONLY);
	if (fd == -1)
		return FALSE;

	ignore_file = strdup(file);

	for(;;)
	{
		char *space = NULL;
		char *line = read_line_fd(fd);

		if (!line)
			break;

		if (strlen(line) == 0)
		{
			myfree(line);
			continue;
		}

		space = strchr(line, ' ');
		if (space)
		{
			*space = 0x00;

			ignores = realloc(ignores, sizeof(ignore) * (n_ignores + 1));
			ignores[n_ignores].channel = strdup(line);

			space++;
			while(*space == ' ')
				space++;

			ignores[n_ignores].nick = strdup(space);
			n_ignores++;
		}
		else
			error_exit(FALSE, "'ignore' (%s) missing parameter: must be 'channel nick'", line);

		myfree(line);
	}

	close(fd);

	return TRUE;
}

void free_ignores(void)
{
	int loop = 0;

	for(loop=0; loop<n_ignores; loop++)
	{
		myfree(ignores[loop].channel);
		myfree(ignores[loop].nick);
	}

	myfree(ignores);
	ignores = NULL;

	n_ignores = 0;
}

void free_favorites(void)
{
	int loop = 0;

	for(loop=0; loop<n_favorite_channels; loop++)
	{
		myfree(favorite_channels[loop].server);
		myfree(favorite_channels[loop].channel);
	}

	myfree(favorite_channels);
	favorite_channels = NULL;

	n_favorite_channels = 0;
}

const char *true_false_str(BOOL val)
{
	if (val)
		return "true";

	return "false";
}

BOOL save_config(BOOL save_channels, char **err_msg)
{
	BOOL failed = FALSE;
	int loop = 0;
	FILE *fh = NULL;

	*err_msg = NULL;

	if (!conf_file)	
	{
		*err_msg = strdup("No configuration file selected");
		return FALSE;
	}

	fh = fopen(conf_file, "wb");

	if (!fh)
	{
		asprintf(err_msg, "Cannot open %s for write access", conf_file);
		return FALSE;
	}

	LOG("default_colorpair=\n");
	failed |= fprintf(fh, "default_colorpair=") == -1;
	emit_colorpair(fh, default_colorpair);
	failed |= fprintf(fh, "\n") == -1;
	LOG("highlight_colorpair=\n");
	failed |= fprintf(fh, "highlight_colorpair=") == -1;
	emit_colorpair(fh, highlight_colorpair);
	failed |= fprintf(fh, "\n") == -1;
	LOG("meta_colorpair=\n");
	failed |= fprintf(fh, "meta_colorpair=") == -1;
	emit_colorpair(fh, meta_colorpair);
	failed |= fprintf(fh, "\n") == -1;
	LOG("error_colorpair=\n");
	failed |= fprintf(fh, "error_colorpair=") == -1;
	emit_colorpair(fh, error_colorpair);
	failed |= fprintf(fh, "\n") == -1;
	LOG("temp_colorpair=\n");
	failed |= fprintf(fh, "temp_colorpair=") == -1;
	emit_colorpair(fh, temp_colorpair);
	failed |= fprintf(fh, "\n") == -1;

	failed |= fprintf(fh, "\n") == -1;

	for(loop=0; loop<n_servers && !failed; loop++)
	{
		int sal = 0, achl = 0;
		server *ps = &server_list[loop];

		failed |= fprintf(fh, "server=%s:%d\n", ps -> server_host, ps -> server_port) == -1;
		failed |= fprintf(fh, "description=%s\n", ps -> description ? ps -> description : ps -> server_host) == -1;
		failed |= fprintf(fh, "name=%s\n", ps -> user_complete_name) == -1;
		failed |= fprintf(fh, "username=%s\n", ps -> username) == -1;
		failed |= fprintf(fh, "password=%s\n", ps -> password) == -1;
		failed |= fprintf(fh, "nickname=%s\n", ps -> nickname) == -1;

		for(achl=0; achl<ps -> n_auto_join && !failed; achl++)
			failed |= fprintf(fh, "auto_join=%s\n", ps -> auto_join[achl]) == -1;

		for(sal=0; sal<ps -> n_send_after_login && !failed; sal++)
			failed |= fprintf(fh, "send_after_login=%s\n", ps -> send_after_login[sal]) == -1;

		if (remember_channels)
		{
			int chl = 0;

			for(chl=0; chl<ps -> n_channels && !failed; chl++)
			{
				BOOL is_auto_join = FALSE;
				const char *cur_channel = ps -> pchannels[chl].channel_name;

				if (!is_channel(cur_channel))
					continue;

				for(achl=0; achl<ps -> n_auto_join; achl++)
				{
					if (strcasecmp(ps -> auto_join[achl], cur_channel) == 0)
					{
						is_auto_join = TRUE;
						break;
					}
				}

				if (!is_auto_join)
					failed |= fprintf(fh, "rejoin=%s\n", cur_channel) == -1;
			}
		}

		failed |= fprintf(fh, ";\n") == -1;
	}

	for(loop=0; loop<n_favorite_channels && !failed; loop++)
	{
		favorite *pf = &favorite_channels[loop];

		failed |= fprintf(fh, "favorite=%s %s\n", pf -> server, pf -> channel) == 0;

		if (failed)
			break;
	}

	if (n_favorite_channels)
		failed |= fprintf(fh, ";\n") == -1;

	failed |= fprintf(fh, "all-colors=%s\n", true_false_str(colors_all)) == -1;
	failed |= fprintf(fh, "meta-colors=%s\n", true_false_str(colors_meta)) == -1;
	failed |= fprintf(fh, "nick-color=%s\n", true_false_str(nick_color)) == -1;
	failed |= fprintf(fh, "partial_highlight_match=%s\n", true_false_str(partial_highlight_match)) == -1;
	failed |= fprintf(fh, "notice_in_serverchannel=%s\n", true_false_str(notice_in_server_channel)) == -1;
	failed |= fprintf(fh, "highlight=%s\n", true_false_str(highlight)) == -1;
	failed |= fprintf(fh, "auto_private_channel=%s\n", true_false_str(auto_private_channel)) == -1;
	failed |= fprintf(fh, "update_clock_at_data=%s\n", true_false_str(update_clock_at_data)) == -1;
	failed |= fprintf(fh, "max_channel_record_lines=%d\n", max_channel_record_lines) == -1;
	failed |= fprintf(fh, "word_cloud_n=%d\n", word_cloud_n) == -1;
	failed |= fprintf(fh, "word_cloud_refresh=%d\n", word_cloud_refresh) == -1;
	failed |= fprintf(fh, "word_cloud_win_height=%d\n", word_cloud_win_height) == -1;
	failed |= fprintf(fh, "word_cloud_min_word_size=%d\n", word_cloud_min_word_size) == -1;
	failed |= fprintf(fh, "store_config_on_exit=%s\n", true_false_str(store_config_on_exit)) == -1;
	failed |= fprintf(fh, "topic_scroll=%s\n", true_false_str(topic_scroll)) == -1;
	failed |= fprintf(fh, "allow_invite=%s\n", true_false_str(allow_invite)) == -1;
	failed |= fprintf(fh, "show_parts=%s\n", true_false_str(show_parts)) == -1;
	failed |= fprintf(fh, "show_mode_changes=%s\n", true_false_str(show_mode_changes)) == -1;
	failed |= fprintf(fh, "show_nick_change=%s\n", true_false_str(show_nick_change)) == -1;
	failed |= fprintf(fh, "show_joins=%s\n", true_false_str(show_joins)) == -1;
	failed |= fprintf(fh, "auto_rejoin=%s\n", true_false_str(auto_rejoin)) == -1;
	failed |= fprintf(fh, "use_nonbasic_colors=%s\n", true_false_str(use_nonbasic_colors)) == -1;
	failed |= fprintf(fh, "inverse_window_heading=%s\n", true_false_str(inverse_window_heading)) == -1;
	failed |= fprintf(fh, "remember_channels=%s\n", true_false_str(remember_channels)) == -1;
	failed |= fprintf(fh, "mark_personal_messages=%s\n", true_false_str(mark_personal_messages)) == -1;
	failed |= fprintf(fh, "space_after_start_marker=%s\n", true_false_str(space_after_start_marker)) == -1;
	failed |= fprintf(fh, "fuzzy_highlight=%s\n", true_false_str(fuzzy_highlight)) == -1;

	if (finger_str)
		failed |= fprintf(fh, "finger_str=%s\n", finger_str) == -1;

	if (dcc_path)
		failed |= fprintf(fh, "dcc_path=%s\n", dcc_path) == -1;

	if (ignore_file)
		failed |= fprintf(fh, "ignore_file=%s\n", ignore_file) == -1;

	if (theme_file)
		failed |= fprintf(fh, "theme=%s\n", theme_file) == -1;

	if (part_message)
		failed |= fprintf(fh, "part_message=%s\n", part_message) == -1;

	if (server_exit_message)
		failed |= fprintf(fh, "server_exit_message=%s\n", server_exit_message) == -1;

	failed |= dump_grep_filters(fh) == FALSE;

	if (failed)
	{
		asprintf(err_msg, "Failed writing to %s: disk full?", conf_file);
		fclose(fh);
		return FALSE;
	}

	asprintf(err_msg, "All fine, written %ld bytes to %s", ftell(fh), conf_file);

	if (fclose(fh))
	{
		asprintf(err_msg, "Problem finalizing %s", conf_file);
		return FALSE;
	}

	return TRUE;
}

int load_config(const char *file)
{
	char *description = NULL, *server_host = NULL, *username = NULL, *password = NULL, *nickname = NULL, *user_complete_name = NULL;
	int server_index = -1;
	int linenr = 0;
	int fd = open(file, O_RDONLY);

	if (fd == -1)
	{
		if (errno == ENOENT)
			return -1;

		error_exit(TRUE, "Cannot open config file %s\n", file);
	}

	conf_file = strdup(file);

	for(;;)
	{
		char *line = read_line_fd(fd);
		char *cmd, *par;
		char *is;

		if (!line)
			break;

		linenr++;

		LOG("line %d: %s\n", linenr, str_or_nothing(line));

		if (strlen(line) == 0)
		{
			myfree(line);
			continue;
		}

		if (line[0] == '#' || line[0] == ';')
		{
			myfree(line);
			continue;
		}

		is = strchr(line, '=');
		if (!is)
			error_exit(FALSE, "config: line %d is missing either command or parameter! (%s)", linenr, line);

		/* find parameter */
		par = is + 1;
		while(*par == ' ')
			par++;

		/* remove spaces around command */
		/* spaces at the start */
		cmd = line;
		while(*cmd == ' ')
			cmd++;
		/* spaces at the end */
		*is = 0x00;
		is--;
		while(*is == ' ')
		{
			*is = 0x00;
			is--;
		}

		LOG("line: %s=%s\n", cmd, par);

		if (strcmp(cmd, "server") == 0 || strcmp(cmd, "send_after_login") == 0 || strcmp(cmd, "auto_join") == 0 || strcmp(cmd, "channel") == 0 || strcmp(cmd, "rejoin") == 0)
		{
			/* all stuff already known? */
			if (server_host)
			{
				if (nickname == NULL)
					error_exit(FALSE, "nickname must be set for %s", server_host);

				server_index = add_server(server_host, username, password, nickname, user_complete_name, description ? description : server_host);
				myfree(server_host);
				server_host = NULL;
				myfree(username);
				myfree(password);
				myfree(nickname);
				myfree(user_complete_name);
				myfree(description);

				username = password = nickname = user_complete_name = description = NULL;
			}
		}

		if (strcmp(cmd, "server") == 0)
		{
			/* new server */
			server_host = strdup(par);
		}
		else if (strcmp(cmd, "favorite") == 0)
		{
			char *space = strchr(par, ' ');
			if (!space)
				error_exit(FALSE, "favorite at line %d: either server or channel missing", linenr);

			*space = 0x00;

			favorite_channels = realloc(favorite_channels, (n_favorite_channels + 1) * sizeof(favorite));
			favorite_channels[n_favorite_channels].server = strdup(par);

			space++;
			while(*space == ' ')
				space++;

			favorite_channels[n_favorite_channels].channel = strdup(space);

			n_favorite_channels++;
		}
		else if (strcmp(cmd, "username") == 0)
			username = strdup(par);
		else if (strcmp(cmd, "password") == 0)
			password = strdup(par);
		else if (strcmp(cmd, "nick") == 0 || strcmp(cmd, "nickname") == 0)
			nickname = strdup(par);
		else if (strcmp(cmd, "name") == 0)
			user_complete_name = strdup(par);
		else if (strcmp(cmd, "description") == 0)
			description = strdup(par);
		else if (strcmp(cmd, "server_exit_message") == 0)
			server_exit_message = strdup(par);
		else if (strcmp(cmd, "part_message") == 0)
			part_message = strdup(par);
		else if (strcmp(cmd, "finger_str") == 0)
			finger_str = strdup(par);
		else if (strcmp(cmd, "mark_personal_messages") == 0)
			mark_personal_messages = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "meta-colors") == 0)
			colors_meta = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "all-colors") == 0)
			colors_all = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "update_clock_at_data") == 0)
			update_clock_at_data = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "nick-color") == 0)
			nick_color = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "use_nonbasic_colors") == 0)
			use_nonbasic_colors = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "inverse_window_heading") == 0)
			inverse_window_heading = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "allow_invite") == 0)
			allow_invite = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "remember_channels") == 0)
			remember_channels = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "auto_rejoin") == 0)
			auto_rejoin = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "space_after_start_marker") == 0)
			space_after_start_marker = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "grep_filter") == 0)
		{
			char *err = NULL;
			char **parts = NULL;
			int n_parts = 0;

			split_string(par, ",", FALSE, &parts, &n_parts);

			if (n_parts != 4)
				error_exit(FALSE, "Parameter missing at line %d\n", linenr);

			if (add_grep_filter(parts[3], parts[0], parts[1], parts[2], &err) == FALSE)
				error_exit(FALSE, "Failed processing grep_filter: %s (line %d)\n", err, linenr);

			free_splitted_string(parts, n_parts);
		}
		else if (strcmp(cmd, "show_parts") == 0)
			show_parts = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "show_mode_changes") == 0)
			show_mode_changes = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "show_nick_change") == 0)
			show_nick_change = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "show_joins") == 0)
			show_joins = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "store_config_on_exit") == 0)
			store_config_on_exit = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "partial_highlight_match") == 0)
			partial_highlight_match = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "topic_scroll") == 0)
			topic_scroll = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "notice_in_serverchannel") == 0)
			notice_in_server_channel = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "highlight") == 0)
			highlight = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "fuzzy_highlight") == 0)
			fuzzy_highlight = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "theme") == 0)
		{
			struct stat status;
			const char *filename = realpath(par, NULL);
			
			if (stat(filename, &status) == -1) 	/* file doesn't exist, look for it under SYSCONFDIR */
			{
				int len = strlen(SYSCONFDIR) + strlen(par) + 2;
				char *theme_path = malloc(len * sizeof(char));
				
				snprintf(theme_path, len, "%s/%s", SYSCONFDIR, par);
				load_theme(theme_path);

				theme_file = theme_path;
			} 
			else
			{
				load_theme(par);

				theme_file = strdup(par);
			}

			myfree(filename);
		}
		else if (strcmp(cmd, "ignore_file") == 0)
		{
			struct stat status;
			const char *filename = realpath(par, NULL);

			if (load_ignore_list(par) == TRUE)
			{
			}
			else if (load_ignore_list(filename) == TRUE)
			{
			}
			else if (stat(filename, &status) == -1) 	/* file doesn't exist, look elsewhere */
			{
				int len = strlen(SYSCONFDIR) + strlen(par) + 2;
				char *ignore_file = malloc(len * sizeof(char));

				/* look for it under SYSCONFDIR */
				snprintf(ignore_file, len, "%s/%s", SYSCONFDIR, par);

				/* look for it under ~/.firc location */
				if (stat(ignore_file, &status) == -1)
					snprintf(ignore_file, len, "%s/%s", dirname(conf_file), par);

				load_ignore_list(ignore_file);

				myfree(ignore_file);
			} 

			myfree(filename);
		}
		else if (strcmp(cmd, "send_after_login") == 0)
		{
			server *ps = &server_list[server_index];

			if (server_index == -1)
				error_exit(FALSE, "send_after_login: you need to define a server first\n");

			ps -> send_after_login = realloc(ps -> send_after_login, sizeof(char *) * (ps -> n_send_after_login + 1));
			ps -> send_after_login[ps -> n_send_after_login] = strdup(par);
			ps -> n_send_after_login++;
		}
		else if (strcmp(cmd, "auto_join") == 0 || strcmp(cmd, "channel") == 0)
		{
			if (server_index == -1)
				error_exit(FALSE, "auto_join: you need to define a server first\n");

			add_autojoin(server_index, par);
		}
		else if (strcmp(cmd, "rejoin") == 0)
			add_channel(server_index, par);
		else if (strcmp(cmd, "auto_private_channel") == 0)
			auto_private_channel = parse_false_true(par, cmd, linenr);
		else if (strcmp(cmd, "dcc_path") == 0)
			dcc_path = strdup(par);
		else if (strcmp(cmd, "default_colorpair") == 0)
			default_colorpair = parse_color_spec(par, linenr, cmd);
		else if (strcmp(cmd, "highlight_colorpair") == 0)
			highlight_colorpair = parse_color_spec(par, linenr, cmd);
		else if (strcmp(cmd, "meta_colorpair") == 0)
			meta_colorpair = parse_color_spec(par, linenr, cmd);
		else if (strcmp(cmd, "error_colorpair") == 0)
			error_colorpair = parse_color_spec(par, linenr, cmd);
		else if (strcmp(cmd, "temp_colorpair") == 0)
			temp_colorpair = parse_color_spec(par, linenr, cmd);
		else if (strcmp(cmd, "check_for_mail") == 0)
			check_for_mail = atoi(par);
		else if (strcmp(cmd, "word_cloud_n") == 0)
		{
			word_cloud_n = atoi(par);
		}
		else if (strcmp(cmd, "word_cloud_refresh") == 0)
		{
			word_cloud_refresh = atoi(par);
			word_cloud_last_refresh = time(NULL);
		}
		else if (strcmp(cmd, "word_cloud_win_height") == 0)
			word_cloud_win_height = atoi(par);
		else if (strcmp(cmd, "max_channel_record_lines") == 0)
			max_channel_record_lines = atoi(par);
		else if (strcmp(cmd, "word_cloud_min_word_size") == 0)
			word_cloud_min_word_size = atoi(par);
		else
		{
			error_exit(FALSE, "'%s=%s' is not understood\n", cmd, par);
		}

		myfree(line);
	}

	close(fd);

	if (server_host)
	{
		if (nickname == NULL)
			error_exit(FALSE, "nickname must be set for %s", server_host);
		add_server(server_host, username, password, nickname, user_complete_name, description);
		myfree(server_host);
		myfree(username);
		myfree(password);
		myfree(nickname);
		myfree(user_complete_name);
		myfree(description);
	}

	LOG("# servers: %d\n", n_servers);

	return 0;
}

BOOL save_ignore_list(void)
{
	int loop = 0;
	BOOL ok = TRUE;
	FILE *fh = NULL;

	if (!ignore_file)
		asprintf(&ignore_file, "%s/%s", dirname(conf_file), ".firc.ignore");

	fh = fopen(ignore_file, "w");
	if (!fh)
		return FALSE;

	for(loop=0; loop<n_ignores; loop++)
	{
		if (fprintf(fh, "%s %s\n", ignores[loop].channel, ignores[loop].nick) < 1)
			ok = FALSE;

		if (!ok)
			LOG("fail: %s\n", strerror(errno));
	}

	fclose(fh);

	return ok;
}

int config_color_str_convert(const char *in, int linenr, const char *subj)
{
	int nr = color_str_convert(in);

	LOG("col convert %s: %d\n", in, nr);

	if (nr >= -1)
		return nr;

	error_exit(FALSE, "Color %s for %s at line %d not recognized", in, subj, linenr);

	return -2;
}

int parse_color_spec(const char *par, int linenr, const char *subj)
{
	char *temp = strdup(par);
	char *komma = strchr(temp, ',');
	const char *fgstr = temp, *bgstr = "default";
	int fg = -1, bg = -1;
	int col = -1;

	if (komma)
	{
		*komma = 0x00;

		bgstr = komma + 1;
	}

	fg = config_color_str_convert(fgstr, linenr, subj);
	bg = config_color_str_convert(bgstr, linenr, subj);

	free(temp);

	if (fg == bg && fg != -1)
		fg = bg = -1;

	LOG("col parse_color_spec(%d,%d)\n", fg, bg);

	col = get_color_ncurses(fg, bg);

	LOG("col parse_color_spec: %d: %d,%d (%s)\n", col, fg, bg, par);

	return col;
}
